# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Download and postprocess ISH dataset from Allen Brain."""
import argparse
import pathlib
import sys

import numpy as np
import PIL


def parse_args():
    """Parse command line arguments.

    Returns
    -------
    args : argparse.Namespace
        The parsed command line arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("gene_name")
    args = parser.parse_args()

    return args


def postprocess_dataset(dataset):
    """Post process given dataset.

    Parameters
    ----------
    dataset : iterable
        List containing dataset, each element being tuple
        (img_id, section_coordinate, img, df)

    Returns
    -------
    dataset_np : np.ndarray
        Array containing gene expressions of the dataset.
    metadata_dict : dict
        Dictionary containing metadata of the dataset.
        Keys are section numbers and image ids.
    """
    metadata_dict = {}
    section_numbers = []
    image_ids = []
    dataset_np = []

    for img_id, section_coordinate, img, df in dataset:
        if section_coordinate is None:
            # In `atldld.sync.download_dataset` if there is a problem during the download
            # the generator returns `(img_id, None, None, None, None)
            # TODO: maybe notify the user somehow?
            continue

        section_numbers.append(section_coordinate // 25)
        image_ids.append(img_id)
        warped_img = df.warp(img, border_mode="constant", c=img[0, 0, :].tolist())
        dataset_np.append(warped_img)

    dataset_np = np.array(dataset_np)

    metadata_dict["section_numbers"] = section_numbers
    metadata_dict["image_ids"] = image_ids
    metadata_dict["image_shape"] = warped_img.shape

    return dataset_np, metadata_dict


def main():
    """Download gene expression dataset."""
    # Imports
    import json

    from atldld.sync import download_dataset
    from atldld.utils import CommonQueries, get_experiment_list_from_gene

    args = parse_args()
    gene = args.gene_name
    # To avoid Decompression Warning
    PIL.Image.MAX_IMAGE_PIXELS = 200000000

    # Download dataset on allen
    for axis in ['sagittal', 'coronal']:

        file_dir = pathlib.Path(f"{axis}/{gene}/")
        if not file_dir.exists():
            file_dir.mkdir(parents=True)
        
        experiment_list = get_experiment_list_from_gene(gene, axis)
        for experiment_id in experiment_list:
            dataset = download_dataset(experiment_id)
            axis = CommonQueries.get_axis(experiment_id)
            dataset_np, metadata_dict = postprocess_dataset(dataset)
            metadata_dict["axis"] = axis

            np.save(file_dir / f"{experiment_id}.npy", dataset_np)
            with open(file_dir / f"{experiment_id}.json", 'w') as f:
                json.dump(metadata_dict, f)


if __name__ == "__main__":
    sys.exit(main())



# SPDX-License-Identifier: MIT
# Source: https://github.com/microsoft/MaskFlownet/tree/5cba12772e2201f0d1c1e27161d224e585334571
import os
import re
import sys

repo_root = os.path.dirname(__file__)

def list_dir(folder, pattern, groups=False):
	pattern = re.compile(pattern)
	for f in os.listdir(folder):
		m = pattern.match(f)
		if m is not None:
			if groups:
				yield (os.path.join(folder, f),) + m.groups()
			else:
				yield os.path.join(folder, f)

def find_log(prefix):
	log_dir = os.path.join(repo_root, 'logs')
	pattern = r'^(%s(.*\d)?)\.log$' % prefix
	rets = list(list_dir(log_dir, pattern, groups=True))
	print(rets)
	if len(rets) > 0 :
		return rets[0][:2]
	else:
		raise ValueError("Not found {}".format(prefix))

def find_checkpoints(run_id):
	weights_dir = os.path.join(repo_root, 'weights')
	pattern = r'^{}.*_(\d+)\.params$'.format(run_id)
	checkpoints = list(sorted(list_dir(weights_dir, pattern, groups=True), key=lambda x : int(x[1])))
	return checkpoints

def read_log(fname):
	val = []
	exp_info = []
	in_start = False
	with open(fname, 'r') as fi:
		for ln in fi:
			p = ln.find('] ')
			items = ln[p + 2:].strip().split(', ')
			try:
				kvs = dict([ item.split('=') for item in items ])
			except:
				pass
			if 'val_epe' in kvs:
				val.append(kvs)
			elif 'start' in kvs:
				exp_info.append(kvs)
				in_start = True
			elif in_start:
				exp_info[-1].update(kvs)
				in_start = False
	return val, exp_info



# SPDX-License-Identifier: MIT
# Source: https://github.com/myungsub/CAIN/tree/2e727d2a07d3f1061f17e2edaa47a7fb3f7e62c5
import math
import numpy as np

import torch
import torch.nn as nn

from .common import *


class Encoder(nn.Module):
    def __init__(self, in_channels=3, depth=3, nf_start=32, norm=False):
        super(Encoder, self).__init__()
        self.device = torch.device('cuda')
        
        nf = nf_start
        relu = nn.LeakyReLU(negative_slope=0.2, inplace=True)

        self.body = nn.Sequential(
            ConvNorm(in_channels, nf * 1, 7, stride=1, norm=norm),
            relu,
            ConvNorm(nf * 1, nf * 2, 5, stride=2, norm=norm),
            relu,
            ConvNorm(nf * 2, nf * 4, 5, stride=2, norm=norm),
            relu,
            ConvNorm(nf * 4, nf * 6, 5, stride=2, norm=norm)
        )
        
        self.interpolate = Interpolation(5, 12, nf * 6, reduction=16, act=relu)

    def forward(self, x1, x2):
        """
        Encoder: Feature Extraction --> Feature Fusion --> Return
        """
        feats1 = self.body(x1)
        feats2 = self.body(x2)

        feats = self.interpolate(feats1, feats2)

        return feats


class Decoder(nn.Module):
    def __init__(self, in_channels=192, out_channels=3, depth=3, norm=False, up_mode='shuffle'):
        super(Decoder, self).__init__()
        self.device = torch.device('cuda')

        relu = nn.LeakyReLU(negative_slope=0.2, inplace=True)

        nf = [in_channels, (in_channels*2)//3, in_channels//3, in_channels//6]
        #nf = [192, 128, 64, 32]
        #nf = [186, 124, 62, 31]
        self.body = nn.Sequential(
            UpConvNorm(nf[0], nf[1], mode=up_mode, norm=norm),
            ResBlock(nf[1], nf[1], norm=norm, act=relu),
            UpConvNorm(nf[1], nf[2], mode=up_mode, norm=norm),
            ResBlock(nf[2], nf[2], norm=norm, act=relu),
            UpConvNorm(nf[2], nf[3], mode=up_mode, norm=norm),
            ResBlock(nf[3], nf[3], norm=norm, act=relu),
            conv7x7(nf[3], out_channels)
        )

    def forward(self, feats):
        out = self.body(feats)
        #out = self.conv_final(out)

        return out


class CAIN_EncDec(nn.Module):
    def __init__(self, depth=3, n_resblocks=3, start_filts=32, up_mode='shuffle'):
        super(CAIN_EncDec, self).__init__()
        self.depth = depth

        self.encoder = Encoder(in_channels=3, depth=depth, norm=False)
        self.decoder = Decoder(in_channels=start_filts*6, depth=depth, norm=False, up_mode=up_mode)

    def forward(self, x1, x2):
        x1, m1 = sub_mean(x1)
        x2, m2 = sub_mean(x2)

        if not self.training:
            paddingInput, paddingOutput = InOutPaddings(x1)
            x1 = paddingInput(x1)
            x2 = paddingInput(x2)

        feats = self.encoder(x1, x2)
        out = self.decoder(feats)

        if not self.training:
            out = paddingOutput(out)

        mi = (m1 + m2)/2
        out += mi

        return out, feats





# SPDX-License-Identifier: MIT
# Source: https://github.com/microsoft/MaskFlownet/tree/5cba12772e2201f0d1c1e27161d224e585334571
from . import layer
from . import MaskFlownet
from . import pipeline

def get_pipeline(network, **kwargs):
	if network == 'MaskFlownet':
		return pipeline.PipelineFlownet(**kwargs)
	else:
		raise NotImplementedError



# SPDX-License-Identifier: MIT
# Source: https://github.com/microsoft/MaskFlownet/tree/5cba12772e2201f0d1c1e27161d224e585334571
import mxnet as mx
import numpy as np
from mxnet import nd, gluon, autograd

from .MaskFlownet import *
from .config import Reader
from .layer import Reconstruction2D, Reconstruction2DSmooth

def build_network(name):
	return eval(name)

def get_coords(img):
	shape = img.shape
	range_x = nd.arange(shape[2], ctx = img.context).reshape(shape = (1, 1, -1, 1)).tile(reps = (shape[0], 1, 1, shape[3]))
	range_y = nd.arange(shape[3], ctx = img.context).reshape(shape = (1, 1, 1, -1)).tile(reps = (shape[0], 1, shape[2], 1))
	return nd.concat(range_x, range_y, dim = 1)


class PipelineFlownet:
	_lr = None

	def __init__(self, ctx, config):
		self.ctx = ctx
		self.network = build_network(getattr(config.network, 'class').get('MaskFlownet'))(config=config)
		self.network.hybridize()
		self.network.collect_params().initialize(init=mx.initializer.MSRAPrelu(slope=0.1), ctx=self.ctx)
		self.trainer = gluon.Trainer(self.network.collect_params(), 'adam', {'learning_rate': 1e-4})
		self.strides = self.network.strides or [64, 32, 16, 8, 4]

		self.scale = self.strides[-1]
		self.upsampler = Upsample(self.scale)
		self.upsampler_mask = Upsample(self.scale)

		self.epeloss = EpeLoss()
		self.epeloss.hybridize()
		self.epeloss_with_mask = EpeLossWithMask()
		self.epeloss_with_mask.hybridize()

		multiscale_weights = config.network.mw.get([.005, .01, .02, .08, .32])
		if len(multiscale_weights) != 5:
			multiscale_weights = [.005, .01, .02, .08, .32]
		self.multiscale_epe = MultiscaleEpe(
			scales = self.strides, weights = multiscale_weights, match = 'upsampling',
			eps = 1e-8, q = config.optimizer.q.get(None))
		self.multiscale_epe.hybridize()

		self.reconstruction = Reconstruction2DSmooth(3)
		self.reconstruction.hybridize()

		self.lr_schedule = config.optimizer.learning_rate.value

	def save(self, prefix):
		self.network.save_parameters(prefix + '.params')
		self.trainer.save_states(prefix + '.states')

	def load(self, checkpoint):
		self.network.load_parameters(checkpoint, ctx=self.ctx)

	def load_head(self, checkpoint):
		self.network.load_head(checkpoint, ctx=self.ctx)

	def fix_head(self):
		self.network.fix_head()

	def set_learning_rate(self, steps):
		i = 0
		while i < len(self.lr_schedule) and steps > self.lr_schedule[i][0]:
			i += 1
		try:
			lr = self.lr_schedule[i][1]
		except IndexError:
			return False
		self.trainer.set_learning_rate(lr)
		self._lr = lr
		return True	

	@property
	def lr(self):
		return self._lr

	def loss(self, pred, occ_masks, labels, masks):
		loss = self.multiscale_epe(labels, masks, *pred)
		return loss
	
	def centralize(self, img1, img2):
		rgb_mean = nd.concat(img1, img2, dim = 2).mean(axis = (2, 3)).reshape((-2, 1, 1))
		return img1 - rgb_mean, img2 - rgb_mean, rgb_mean

	def train_batch(self, img1, img2, label, geo_aug, color_aug, mask = None):
		losses = []
		epes = []
		batch_size = img1.shape[0]
		if mask is None:
			mask = np.full(shape = (batch_size, 1, 1, 1), fill_value = 255, dtype = np.uint8)
		img1, img2, label, mask = map(lambda x : gluon.utils.split_and_load(x, self.ctx), (img1, img2, label, mask))
		
		with autograd.record():
			for img1s, img2s, labels, masks in zip(img1, img2, label, mask):
				img1s, img2s, labels, masks = img1s / 255.0, img2s / 255.0, labels.astype("float32", copy=False), masks / 255.0
				img1s, img2s, labels, masks = geo_aug(img1s, img2s, labels, masks)
				img1s, img2s = color_aug(img1s, img2s)
				img1s, img2s, rgb_mean = self.centralize(img1s, img2s)
				pred, occ_masks, warpeds = self.network(img1s, img2s)
				
				labels = labels.flip(axis = 1)
				loss = self.loss(pred, occ_masks, labels, masks)
				epe = self.epeloss_with_mask(self.upsampler(pred[-1]), labels, masks)

				losses.append(loss)
				epes.append(epe)
		
		for loss in losses:
			loss.backward()
		self.trainer.step(batch_size)
		return {"epe": np.mean(np.concatenate([epe.asnumpy() for epe in epes]))}

	def do_batch_mx(self, img1, img2, resize = None):
		''' do a batch of samples range [0,1] with network preprocessing and padding
		'''
		img1, img2, _ = self.centralize(img1, img2)
		shape = img1.shape
		if resize is None:
			pad_h = (64 - shape[2] % 64) % 64
			pad_w = (64 - shape[3] % 64) % 64
		else:
			pad_h = resize[0] - shape[2]
			pad_w = resize[1] - shape[3]
		if pad_h != 0 or pad_w != 0:
			img1 = nd.contrib.BilinearResize2D(img1, height = shape[2] + pad_h, width = shape[3] + pad_w)
			img2 = nd.contrib.BilinearResize2D(img2, height = shape[2] + pad_h, width = shape[3] + pad_w)
		pred, flows, warpeds = self.network(img1, img2)
		return pred, flows, warpeds

	def do_batch(self, img1, img2, label = None, mask = None, resize = None):
		shape = img1.shape
		flows, occ_masks, _ = self.do_batch_mx(img1, img2, resize = resize)
		flow = self.upsampler(flows[-1])
		occ_mask = self.upsampler(occ_masks[0])
		if shape[2] != flow.shape[2] or shape[3] != flow.shape[3]:
			flow = nd.contrib.BilinearResize2D(flow, height = shape[2], width = shape[3]) * nd.array(
				[shape[d] / flow.shape[d] for d in (2, 3)], ctx = flow.context).reshape((1, 2, 1, 1))
			occ_mask = nd.contrib.BilinearResize2D(occ_mask, height = shape[2], width = shape[3])
		warp = self.reconstruction(img2, flow)
		epe = None
		if label is not None and mask is not None:
			epe = self.epeloss_with_mask(flow, label, mask)
		return flow, occ_mask, warp, epe

	def validate(self, img1, img2, label, mask = None, batch_size = 1, resize = None, return_type = 'epe'):
		''' validate the whole dataset
		'''
		np_epes = []
		size = len(img1)
		bs = batch_size
		if mask is None:
			mask = [np.full(shape = (1, 1, 1), fill_value = 255, dtype = np.uint8)] * size
		for j in range(0, size, bs):
			batch_img1 = img1[j: j + bs]
			batch_img2 = img2[j: j + bs]
			batch_label = label[j: j + bs]
			batch_mask = mask[j: j + bs]

			batch_img1 = np.transpose(np.stack(batch_img1, axis=0), (0, 3, 1, 2))
			batch_img2 = np.transpose(np.stack(batch_img2, axis=0), (0, 3, 1, 2))
			batch_label = np.transpose(np.stack(batch_label, axis=0), (0, 3, 1, 2))
			batch_mask = np.transpose(np.stack(batch_mask, axis=0), (0, 3, 1, 2))

			def Norm(x):
				return nd.sqrt(nd.sum(nd.square(x), axis = 1, keepdims = True))

			batch_epe = []
			ctx = self.ctx[ : min(len(batch_img1), len(self.ctx))]
			nd_img1, nd_img2, nd_label, nd_mask = map(lambda x : gluon.utils.split_and_load(x, ctx, even_split = False), (batch_img1, batch_img2, batch_label, batch_mask))
			for img1s, img2s, labels, masks in zip(nd_img1, nd_img2, nd_label, nd_mask):
				img1s, img2s, labels, masks = img1s / 255.0, img2s / 255.0, labels.astype("float32", copy=False), masks / 255.0
				labels = labels.flip(axis = 1)
				flows, _, _, epe = self.do_batch(img1s, img2s, labels, masks, resize = resize)

				# calculate the metric for kitti dataset evaluation
				if return_type is not 'epe':
					eps = 1e-8
					epe = ((Norm(flows - labels) > 3) * ((Norm(flows - labels) / (Norm(labels) + eps)) > 0.05) * masks).sum(axis=0, exclude=True) / masks.sum(axis=0, exclude=True)
				
				batch_epe.append(epe)
			np_epes.append(np.concatenate([epe.asnumpy() for epe in batch_epe]))
		
		return np.mean(np.concatenate(np_epes, axis = 0), axis = 0)

	def predict(self, img1, img2, batch_size, resize = None):
		''' predict the whole dataset
		'''
		size = len(img1)
		bs = batch_size
		for j in range(0, size, bs):
			batch_img1 = img1[j: j + bs]
			batch_img2 = img2[j: j + bs]

			batch_img1 = np.transpose(np.stack(batch_img1, axis=0), (0, 3, 1, 2))
			batch_img2 = np.transpose(np.stack(batch_img2, axis=0), (0, 3, 1, 2))

			batch_flow = []
			batch_occ_mask = []
			batch_warped = []

			ctx = self.ctx[ : min(len(batch_img1), len(self.ctx))]
			nd_img1, nd_img2 = map(lambda x : gluon.utils.split_and_load(x, ctx, even_split = False), (batch_img1, batch_img2))
			for img1s, img2s in zip(nd_img1, nd_img2):
				img1s, img2s = img1s / 255.0, img2s / 255.0
				flow, occ_mask, warped, _ = self.do_batch(img1s, img2s, resize = resize)
				batch_flow.append(flow)
				batch_occ_mask.append(occ_mask)
				batch_warped.append(warped)
			
			flow = np.concatenate([x.asnumpy() for x in batch_flow])
			occ_mask = np.concatenate([x.asnumpy() for x in batch_occ_mask])
			warped = np.concatenate([x.asnumpy() for x in batch_warped])
			
			flow = np.transpose(flow, (0, 2, 3, 1))
			flow = np.flip(flow, axis = -1)
			occ_mask = np.transpose(occ_mask, (0, 2, 3, 1))
			warped = np.transpose(warped, (0, 2, 3, 1))
			for k in range(len(flow)):
				yield flow[k], occ_mask[k], warped[k]



# SPDX-License-Identifier: MIT
# Source: https://github.com/microsoft/MaskFlownet/tree/5cba12772e2201f0d1c1e27161d224e585334571
import datetime
import os
import shutil

class FileLog:
	__timezone__ = 8

	def __init__(self, path, name=None, screen=False):
		if os.path.exists(path):
			shutil.copy(path, path + '.bak')
		self._path = path
		self.f = open(path, 'a')
		self.screen = screen

	@classmethod
	def _localtime(cls):
		return datetime.datetime.utcnow() + datetime.timedelta(hours=cls.__timezone__)

	def _timestamp(self):
		return (self._localtime()).strftime('[%Y/%m/%d %H:%M:%S] ')
	
	def log(self, msg, end='\n'):
		self.f.write(self._timestamp() + msg + end)
		self.f.flush()
		if self.screen:
			print(self._timestamp() + msg + end, end='')

	def close(self):
		self.f.close()



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock

import numpy as np
import pytest
import torch

from atlinter.data import GeneDataset
from atlinter.pair_interpolation import (
    AntsPairInterpolationModel,
    CAINPairInterpolationModel,
    GeneInterpolate,
    LinearPairInterpolationModel,
    PairInterpolate,
    PairInterpolationModel,
    RIFEPairInterpolationModel,
)


class DummyModel(PairInterpolationModel):
    def interpolate(self, img1, img2):
        raise NotImplementedError


def test_pair_interpolation_model():
    # Test input
    img1 = np.random.rand()
    img2 = np.random.rand()
    interpolated_images = np.random.rand()

    # Instantiate model
    model = DummyModel()

    # Tests
    ret1, ret2 = model.before_interpolation(img1, img2)
    ret = model.after_interpolation(interpolated_images)
    assert ret1 is img1
    assert ret2 is img2
    assert ret is interpolated_images


@pytest.mark.parametrize(
    ("n_repeat", "n_expected_interpolations"),
    (
        (1, 1),
        (2, 3),
        (3, 7),
        (10, 1023),
    ),
)
@pytest.mark.parametrize("shape", [(10, 10), (10, 10, 3)])
def test_pair_interpolate(n_repeat, n_expected_interpolations, shape):
    # Test input
    img1 = np.random.rand(*shape)
    img2 = np.random.rand(*shape)

    # Interpolation model mock
    model = mock.Mock(spec=PairInterpolationModel)
    dummy_model = DummyModel()
    model.before_interpolation = mock.Mock(wraps=dummy_model.before_interpolation)
    model.interpolate = mock.Mock(return_value=np.random.rand(*shape))
    model.after_interpolation = mock.Mock(wraps=dummy_model.after_interpolation)

    # Run interpolation
    interpolate = PairInterpolate().repeat(n_repeat)
    interpolated = interpolate(img1, img2, model)

    # Check results
    assert isinstance(interpolated, np.ndarray)
    assert len(interpolated.shape) == len(shape) + 1
    assert interpolated.shape[1:] == img1.shape
    assert len(interpolated) == n_expected_interpolations
    model.before_interpolation.assert_called_once()
    assert model.interpolate.call_count == n_expected_interpolations
    model.after_interpolation.assert_called_once()


@pytest.mark.parametrize("shape", [(10, 10), (10, 10, 3)])
def test_linear_pair_interpolation_model(shape):
    # Test input
    img1 = np.random.rand(*shape)
    img2 = np.random.rand(*shape)

    # Test instances
    model = LinearPairInterpolationModel()
    interpolate = PairInterpolate()

    # Tests
    interpolated = interpolate(img1, img2, model)
    assert interpolated.shape == (1, *shape)
    assert np.allclose(interpolated[0], np.mean([img1, img2], axis=0))


@pytest.mark.parametrize("shape", [(10, 10), (10, 10, 3)])
def test_ants_pair_interpolation_model(shape):
    # Test input
    img1 = np.random.rand(*shape)
    img2 = np.random.rand(*shape)
    img_mid = np.random.rand(*shape)

    # Test instances
    register_fn = mock.Mock(return_value=np.random.rand())
    transform_fn = mock.Mock(return_value=img_mid)
    model = AntsPairInterpolationModel(register_fn, transform_fn)
    interpolate = PairInterpolate()

    # Tests
    interpolated = interpolate(img1, img2, model)
    assert interpolated.shape == (1, *shape)
    register_fn.assert_called_once()
    transform_fn.assert_called_once()


@pytest.mark.parametrize("shape", [(10, 10), (10, 10, 3)])
def test_cain_pair_interpolation_model(shape):
    # Test input
    img1 = np.random.rand(*shape)
    img2 = np.random.rand(*shape)
    img_mid = torch.rand(1, 3, *shape[0:2])

    # Test instances
    cain_model = mock.Mock(return_value=(img_mid, None))
    model = CAINPairInterpolationModel(cain_model)
    interpolate = PairInterpolate()

    # Tests
    interpolated = interpolate(img1, img2, model)
    assert interpolated.shape == (1, *shape)
    cain_model.assert_called_once()


@pytest.mark.parametrize("shape", [(10, 10), (10, 10, 3)])
def test_rife_pair_interpolation_model(shape):
    # Test input
    img1 = np.random.rand(*shape)
    img2 = np.random.rand(*shape)
    img_mid = torch.rand(1, 3, *shape[0:2])

    # Test instances
    rife_model = mock.Mock()
    rife_model.inference = mock.Mock(return_value=img_mid)
    rife_device = None
    model = RIFEPairInterpolationModel(rife_model, rife_device)
    interpolate = PairInterpolate()

    # Tests
    interpolated = interpolate(img1, img2, model)
    assert interpolated.shape == (1, *shape)
    rife_model.inference.assert_called_once()


class FakeModel(PairInterpolationModel):
    def interpolate(self, img1, img2):
        img_middle = (img1 + img2) / 2
        return img_middle


def fake_gene_data(axis, slice_shape, volume_shape):
    n_known_slices = 3
    gene_volume = np.ones((n_known_slices, *slice_shape))

    for i in range(n_known_slices):
        gene_volume[i] = (i + 1) * gene_volume[i]

    section_numbers = [10 + i * 8 for i in range(n_known_slices)]

    return GeneDataset(gene_volume, section_numbers, axis, volume_shape)


class TestGeneInterpolate:
    def test_get_n_repeat(self):
        n_rep = [0, 0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4]
        for diff, n in enumerate(n_rep):
            assert GeneInterpolate.get_n_repeat(diff) == n

    @pytest.mark.parametrize(
        ("left", "right", "n_repeat", "expected_output"),
        [
            (10, 11, 0, []),
            (10, 20, 1, np.array([15])),
            (10, 12, 2, np.array([10.5, 11, 11.5])),
            (10, 18, 3, np.arange(11, 18)),
            (10, 15, 3, [10.625, 11.25, 11.875, 12.5, 13.125, 13.75, 14.375]),
        ],
    )
    def test_get_predicted_section_numbers(
        self, left, right, n_repeat, expected_output
    ):
        """Check that predicted section numbers are good"""
        assert np.all(
            GeneInterpolate.get_predicted_section_numbers(left, right, n_repeat)
            == expected_output
        )

    @pytest.mark.parametrize(
        "axis",
        [
            "coronal",
            "sagittal",
        ],
    )
    @pytest.mark.parametrize("volume_shape", [(30, 20, 40), (32, 18, 29)])
    @pytest.mark.parametrize("rgb", [True, False])
    def test_predict_slice(self, axis, volume_shape, rgb):
        """Check that one can predict one slice."""
        if axis == "coronal":
            shape = (volume_shape[1], volume_shape[2])
        else:
            shape = (volume_shape[0], volume_shape[1])

        if rgb:
            volume_shape = (*volume_shape, 3)
            shape = (*shape, 3)

        gene_data = fake_gene_data(axis, shape, volume_shape)
        gene_interpolate = GeneInterpolate(gene_data, FakeModel())
        prediction = gene_interpolate.predict_slice(5)
        assert isinstance(prediction, np.ndarray)
        assert np.all(prediction.shape == shape)

        # Slice to predict is < first known gene slice, copy the first one
        assert np.all(gene_interpolate.predict_slice(5) == np.ones(shape) * 1)
        # Slice to predict is > last known gene slice, copy the last one
        assert np.all(gene_interpolate.predict_slice(100) == np.ones(shape) * 3)
        # Result of the first iteration
        assert np.all(gene_interpolate.predict_slice(14) == np.ones(shape) * 1.5)
        # Result of the second iteration
        assert np.all(gene_interpolate.predict_slice(12) == np.ones(shape) * 1.25)
        assert np.all(gene_interpolate.predict_slice(16) == np.ones(shape) * 1.75)
        # Result of the third iteration
        assert np.all(gene_interpolate.predict_slice(11) == np.ones(shape) * 1.125)
        assert np.all(gene_interpolate.predict_slice(17) == np.ones(shape) * 1.875)

    @pytest.mark.parametrize(
        "axis",
        [
            "coronal",
            "sagittal",
        ],
    )
    @pytest.mark.parametrize("volume_shape", [(30, 20, 40), (32, 18, 29)])
    @pytest.mark.parametrize("rgb", [True, False])
    def test_get_all_predictions(self, axis, volume_shape, rgb):
        """Check that one can predict entire volume."""
        if axis == "coronal":
            shape = (volume_shape[1], volume_shape[2])
        else:
            shape = (volume_shape[0], volume_shape[1])

        if rgb:
            volume_shape = (*volume_shape, 3)
            shape = (*shape, 3)

        gene_data = fake_gene_data(axis, shape, volume_shape)
        gene_interpolate = GeneInterpolate(gene_data, FakeModel())
        (
            all_interpolated_images,
            all_predicted_section_numbers,
        ) = gene_interpolate.get_all_interpolation()
        assert isinstance(all_interpolated_images, np.ndarray)
        assert isinstance(all_predicted_section_numbers, np.ndarray)
        assert np.all(all_predicted_section_numbers.shape == (14,))
        # 14 = 2 intervals * 7 interpolated images per interval
        assert np.all(all_interpolated_images.shape == (14, *shape))

    @pytest.mark.parametrize(
        "axis",
        [
            "coronal",
            "sagittal",
        ],
    )
    @pytest.mark.parametrize("volume_shape", [(30, 20, 40), (32, 18, 29)])
    @pytest.mark.parametrize("rgb", [True, False])
    def test_predict_volume(self, axis, volume_shape, rgb):
        """Check that one can predict entire volume."""
        if axis == "coronal":
            shape = (volume_shape[1], volume_shape[2])
        else:
            shape = (volume_shape[0], volume_shape[1])

        if rgb:
            volume_shape = (*volume_shape, 3)
            shape = (*shape, 3)

        gene_data = fake_gene_data(axis, shape, volume_shape)
        gene_interpolate = GeneInterpolate(gene_data, FakeModel())
        predicted_volume = gene_interpolate.predict_volume()
        assert isinstance(predicted_volume, np.ndarray)
        assert np.all(predicted_volume.shape == volume_shape)

        if axis == "sagittal":
            if rgb:
                predicted_volume = np.transpose(predicted_volume, (2, 0, 1, 3))
            else:
                predicted_volume = np.transpose(predicted_volume, (2, 0, 1))

        assert np.all(np.unique(predicted_volume[0:10]) == np.array([1]))
        assert np.all(predicted_volume[36:] == np.array([3]))

    def test_predict_volume_wrong_axis(self):

        gene_data = GeneDataset(
            np.zeros([1, 10, 10, 3]), [], "fake_axis", (10, 10, 10, 3)
        )
        gene_interpolate = GeneInterpolate(gene_data, FakeModel())
        with pytest.raises(ValueError):
            _ = gene_interpolate.predict_volume()
